from node import *
import os
from parse import parsestr
# from pwifile import pathpush, pathpop
from common import warning, error
import sys

def _eval_tree (node):
    if not node:
        return

    to_pop = False

    if isinstance (node, TagNode):
        for k in node.attrs.keys ():
            root = EmptyNode ()
            root.child = node.attrs[k]
            _eval_tree (root)
            node.attrs[k] = root.child
            node.attrs[k].parent = None

    _eval_tree (node.child)
    _eval_tree (node.next)

    # We're Merging the Content Nodes.
    # This is important since some macros rely on that.
    if isinstance (node, ContentNode) and isinstance (node.next, ContentNode) and node.next.indent == -1:
        mergedNode = ContentNode (node.content + node.next.content)
        node.next.pop ()
        indent = node.indent
        mergedNode.replace (node)
        mergedNode.indent = indent

    if to_pop:
        node.pop ()


def eval_tree (root=None, exec_defered=True):
    '''
        Create a new tree from the parsed one, executing the command nodes.
        Also, execute the deferred functions.
        The resulting Tree will only be composed of EmptyNode, ContentNode and TagNodes
        and is meant to be processed by the output module's functions.
    '''
    make_root = False
    if not root.parent:
        make_root = True
        fake_root = EmptyNode ()
        fake_root.child = root
    else:
        fake_root = root # not very pretty.

    _eval_tree (fake_root)

    if make_root:
        root = fake_root.child
        root.parent = None
    return root

def eval_string (string=None):
    '''
        Parse a string then evaluate it.
    '''
    return eval_tree (parsestr (string))




#!/usr/bin/env python

from lineparser import parser, set_current_pwiline
from node import *

def parselines (lines):
    '''
        Parse a whole line set (eg. a file)
        On each line, invoke the line parser to get the resulting tree.
        All these trees are then assembled to make the tree of the whole file.

        @param lines : An array of PwiLine
        @return      : The root node of the resulting tree.
    '''
    root = EmptyNode () # This node is meant to be discarded.
    stack = [root]

    # This is to avoid nasty errors when somebody deems intelligent
    # to add spaces on the first element.
    if lines and lines[0].indent != 0:
        lines[0].indent = 0

    for line in lines:
        # We first get the root node of the current line.
        set_current_pwiline (line)
        node = parser.parse (line.content)
        if not node:
            node = EmptyNode ()
        node.indent = line.indent

        # We unstack our tree since the indentation decreased.
        while line.indent < stack[0].indent:
            stack.pop (0)

        if line.indent > stack[0].indent: # We are increasing the level.
            if stack[0].last.child:
                stack[0].last.child.last.next = node
            else:
                stack[0].last.child = node
            stack.insert (0, node)
        elif line.indent == stack[0].indent:
            stack[0].last.next = node

    res = root.child
    if res: # Resulting tree might be None
        res.parent = None
    return res

def parsestr (line):
    '''
        Parse a line or an array of lines
        @param line : A string or an array of string.
             if line if a string, it will be split by \n.
        @return     : The root node of the resulting tree
    '''
    rawlines = line
    if isinstance (line, str) or isinstance (line, unicode):
        rawlines = line.split ('\n')
    lines = []
    for l in rawlines:
        if l.strip () != "":
            lines.append (PwiLine (l, -1, "<str>"))
    return parselines (lines)




from node import *
import os
from common import warning, error
import sys

###################### TREE WALK GENERATOR #######################

ENTER = 0
LEAVE = 1

def gen_treewalk (node):
    if not node:
        return
    stack = [node]

    while stack:
        yield ENTER, node
        if node.child:
            stack.append (node)
            node = node.child
            continue

        if node.next:
            yield LEAVE, node
            node = node.next
            continue

        while stack and not node.next:
            yield LEAVE, node
            node = stack.pop ()
        if not stack:
            break
        yield LEAVE, node
        node = node.next

##########################################################

from cStringIO import StringIO
import traceback

# Write the opening of a TagNode
def tag_opening (node):
    id = ""
    if node.id:
        id = ' id=\"%s\"' % node.id

    classes = ""
    # If there are classes, compute the class attribute
    if node.classes:
        first = True
        for c in node.classes:
            if not first:
                classes += " %s" % c
            else:
                classes += c
                first = False
        classes = " class=\"%s\"" % (classes)

    attrs = ""
    if node.attrs:
        for k in node.attrs:
            buf = StringIO ()
            output_flat (node.attrs[k], buf)
            attrs += " %s=%s" % (k, buf.getvalue ())
    if node.self_closing and not node.child:
        return "<%s%s%s%s/>" % (node.tag, id, classes, attrs)
    return "<%s%s%s%s>" % (node.tag, id, classes, attrs)

# The XML closing part of a tag.
def tag_closing (node):
    if node.self_closing and not node.child:
        return ""
    return "</%s>" % (node.tag)

###################### REAL OUTPUT ########################

from evaluate import eval_tree

def tree_to_string (root, evaluator):
    return output_flat_to_string (eval_tree (root, evaluator))

def output_flat_to_string (root):
    buf = StringIO ()
    output_flat (root, buf)
    return buf.getvalue ()

def output_indented (root, out=sys.stdout):
    last_newline = True

    for status, node in gen_treewalk (root):
        istag = isinstance (node, TagNode)
        isempty = isinstance (node, EmptyNode)
        iscontent = isinstance (node, ContentNode)

        if status == ENTER:
            if node.indent != -1:
                if not isempty and not last_newline:
                    out.write ('\n')
                if not isempty:
                    out.write (' ' * node.indent)

            if istag:
                out.write (tag_opening (node))
            if iscontent:
                out.write (node.content)

            if not isempty:
                if node.child and node.child.indent > -1:
                    out.write ('\n')
                    last_newline = True
                else:
                    last_newline = False

        else: # We're leaving the node ! -- its content already processed
            if istag:
                if node.child and node.child.indent > -1:
                    if not last_newline:
                        out.write ('\n')
                    out.write (' ' * node.indent)
                out.write (tag_closing (node))
            #if not iscmd and node.next and node.next.indent != -1:
            #    out.write ('\n')

    out.write ('\n')

def output_flat (root, out=sys.stdout):

    for status, node in gen_treewalk (root):
        istag = isinstance (node, TagNode)
        iscontent = isinstance (node, ContentNode)

        if status == ENTER:
            if istag:
                out.write (tag_opening (node))
            if iscontent:
                out.write (node.content)

        else: # We're leaving the node ! -- its content already processed
            if istag:
                out.write (tag_closing (node))
            if node.next and node.next.indent != -1:
                out.write (' ')





# This is the library !

import common, preproc, parse, output, evaluate
import sys
from cStringIO import StringIO
import string

from jinja2 import Environment, FileSystemLoader, Template
from jinja2.loaders import split_template_path, open_if_exists, TemplateNotFound

from jinjaext import filters, AbspathExtension

from os import path, getcwd, symlink, remove, chdir
from os.path import abspath, dirname, expanduser

class PwilangLoader (FileSystemLoader):
    def __init__ (self, *args, **kwargs):
        self.with_pwilang = kwargs.pop ('with_pwilang', True)
        super (PwilangLoader, self).__init__ (*args, **kwargs)

    def get_source(self, environment, template):
        pieces = split_template_path(template)
        for searchpath in self.searchpath:
            filename = path.join(searchpath, *pieces)
            f = open_if_exists(filename)
            if f is None:
                continue

            try:
                contents = f.read().decode(self.encoding)
            finally:
                f.close()

            mtime = path.getmtime(filename)
            def uptodate():
                try:
                    return path.getmtime(filename) == mtime
                except OSError:
                    return False

            if self.with_pwilang:
                lines = preproc.get_lines_from_string (contents)

                root = None
                if lines:
                    root = evaluate.eval_tree (parse.parselines (lines))
                    buf = StringIO ()
                    output.output_indented (root, buf)
                    contents = buf.getvalue ()
                    buf.close ()

                    error = common.error_flag
                    errors = StringIO ()

                    common.flush_errors (errors)
                    err = errors.getvalue ()
                    errors.close ()

                    if error:
                        raise Exception (err)

            return contents, filename, uptodate
        raise TemplateNotFound(template)

here = abspath (dirname (__file__))
default_dirs = [getcwd (), path.join (getcwd (), 'templates'), expanduser ("~/.pwilang/templates"), path.join (here, 'templates')]

jinja_env = None
def set_env (dirs=default_dirs):
    global jinja_env
    loader = PwilangLoader (dirs)
    jinja_env = Environment (loader=loader, extensions=[AbspathExtension])

def process (iterable, out=sys.stdout, ctx={}, with_pwilang=True, with_jinja2=True):
    if isinstance (iterable, str) or isinstance (iterable, unicode):
        iterable = iterable.split ('\n')

    contents = ""

    # PWILANG PART
    if with_pwilang:
        lines = preproc.get_lines_from_iterable (iterable)

        root = None
        if lines:
            root = evaluate.eval_tree (parse.parselines (lines))
            buf = StringIO ()
            output.output_indented (root, buf)
            contents = buf.getvalue ()
            buf.close ()

            error = common.error_flag
            errors = StringIO ()

            if out == sys.stdout:
                common.flush_errors ()
            else:
                common.flush_errors (errors)
                err = errors.getvalue ()
                errors.close ()

            if error:
                raise Exception (err)
    else:
        contents = string.join (iterable) # \n ?

    if with_jinja2:
        global jinja_env
        if not jinja_env:
            set_env ()
        env = jinja_env

        if not with_pwilang:
            loader = FileSystemLoader (default_dirs, with_pwilang=False)
            env = Environment (loader=loader, extensions=[AbspathExtension])

        for key, value in filters.items ():
            env.filters[key] = value

        contents = unicode (contents, 'utf-8')
        tmpl = env.from_string (contents)
        contents = tmpl.render (ctx)

    out.write (contents.encode ('utf-8'))

### PDF SECTION

import re
re_wkhtmloptions = re.compile ("<!--wkhtmltopdf (?P<options>.*?)-->")
wkhtmltopdf = "wkhtmltopdf"
from subprocess import call
import shlex

def create_pdf (filename, pdfname, headers=True):
    """ Creates a PDF file using shlex """
    f = open (filename, 'r')
    htmlcontents = f.read ()
    f.close ()
    match = re_wkhtmloptions.search (htmlcontents)

    options = ""
    if match:
        options = match.group ('options')

    header_path = abspath ('%s_header.html' % (filename))
    try:
        symlink (abspath (filename), header_path)
    except:
        pass

    footer_path = abspath ('%s_footer.html' % (filename))
    try:
        symlink (abspath (filename), footer_path)
    except:
        pass

    if headers:
        cmd = "%(wkhtmltopdf)s %(options)s %(source)s --header-html %(header)s --footer-html %(footer)s %(pdfname)s" % dict (wkhtmltopdf=wkhtmltopdf, options=options, source=filename, pdfname=pdfname, header=header_path, footer=footer_path)
    else:
        cmd = "%(wkhtmltopdf)s %(options)s %(source)s %(pdfname)s" % dict (wkhtmltopdf=wkhtmltopdf, options=options, source=filename, pdfname=pdfname, header=header_path, footer=footer_path)

    args = shlex.split (cmd.encode ('utf-8'))
    call (args)

    remove (header_path)
    remove (footer_path)



# Yapps 2.0 Runtime
#
# This module is needed to run generated parsers.

from string import join, count, find, rfind
import re

class SyntaxError(Exception):
    """When we run into an unexpected token, this is the exception to use"""
    def __init__(self, pos=-1, msg="Bad Token"):
        Exception.__init__(self)
        self.pos = pos
        self.msg = msg
    def __repr__(self):
        if self.pos < 0: return "#<syntax-error>"
        else: return "SyntaxError[@ char %s: %s]" % (repr(self.pos), self.msg)

class NoMoreTokens(Exception):
    """Another exception object, for when we run out of tokens"""
    pass

class Scanner:
    def __init__(self, patterns, ignore, input):
        """Patterns is [(terminal,regex)...]
        Ignore is [terminal,...];
        Input is a string"""
        self.tokens = []
        self.restrictions = []
        self.input = input
        self.pos = 0
        self.ignore = ignore
        # The stored patterns are a pair (compiled regex,source
        # regex).  If the patterns variable passed in to the
        # constructor is None, we assume that the class already has a
        # proper .patterns list constructed
        if patterns is not None:
            self.patterns = []
            for k, r in patterns:
                self.patterns.append( (k, re.compile(r)) )
        
    def token(self, i, restrict=0):
        """Get the i'th token, and if i is one past the end, then scan
        for another token; restrict is a list of tokens that
        are allowed, or 0 for any token."""
        if i == len(self.tokens): self.scan(restrict)
        if i < len(self.tokens):
            # Make sure the restriction is more restricted
            if restrict and self.restrictions[i]:
                for r in restrict:
                    if r not in self.restrictions[i]:
                        raise NotImplementedError("Unimplemented: restriction set changed")
            return self.tokens[i]
        raise NoMoreTokens()
    
    def __repr__(self):
        """Print the last 10 tokens that have been scanned in"""
        output = ''
        for t in self.tokens[-10:]:
            output = '%s\n  (@%s)  %s  =  %s' % (output,t[0],t[2],repr(t[3]))
        return output
    
    def scan(self, restrict):
        """Should scan another token and add it to the list, self.tokens,
        and add the restriction to self.restrictions"""
        # Keep looking for a token, ignoring any in self.ignore
        while 1:
            # Search the patterns for the longest match, with earlier
            # tokens in the list having preference
            best_match = -1
            best_pat = '(error)'
            for p, regexp in self.patterns:
                # First check to see if we're ignoring this token
                if restrict and p not in restrict and p not in self.ignore:
                    continue
                m = regexp.match(self.input, self.pos)
                if m and len(m.group(0)) > best_match:
                    # We got a match that's better than the previous one
                    best_pat = p
                    best_match = len(m.group(0))
                    break # Added to prevent options discarding. (pwipwi)
                    
            # If we didn't find anything, raise an error
            if best_pat == '(error)' and best_match < 0:
                msg = "Bad Token"
                if restrict:
                    msg = "Trying to find one of "+join(restrict,", ")
                raise SyntaxError(self.pos, msg)

            # If we found something that isn't to be ignored, return it
            if best_pat not in self.ignore:
                # Create a token with this data
                token = (self.pos, self.pos+best_match, best_pat,
                         self.input[self.pos:self.pos+best_match])
                self.pos = self.pos + best_match
                # Only add this token if it's not in the list
                # (to prevent looping)
                if not self.tokens or token != self.tokens[-1]:
                    self.tokens.append(token)
                    self.restrictions.append(restrict)
                return
            else:
                # This token should be ignored ..
                self.pos = self.pos + best_match

class Parser:
    def __init__(self, scanner):
        self._scanner = scanner
        self._pos = 0
        
    def _peek(self, *types):
        """Returns the token type for lookahead; if there are any args
        then the list of args is the set of token types to allow"""
        tok = self._scanner.token(self._pos, types)
        return tok[2]
        
    def _scan(self, type):
        """Returns the matched text, and moves to the next token"""
        tok = self._scanner.token(self._pos, [type])
        if tok[2] != type:
            raise SyntaxError(tok[0], 'Trying to find '+type)
        self._pos = 1+self._pos
        return tok[3]



def print_error(input, err, scanner):
    """This is a really dumb long function to print error messages nicely."""
    p = err.pos
    # Figure out the line number
    line = count(input[:p], '\n')
    print err.msg+" on line "+repr(line+1)+":"
    # Now try printing part of the line
    text = input[max(p-80, 0):p+80]
    p = p - max(p-80, 0)

    # Strip to the left
    i = rfind(text[:p], '\n')
    j = rfind(text[:p], '\r')
    if i < 0 or (0 <= j < i): i = j
    if 0 <= i < p:
        p = p - i - 1
        text = text[i+1:]

    # Strip to the right
    i = find(text,'\n', p)
    j = find(text,'\r', p)
    if i < 0 or (0 <= j < i): i = j
    if i >= 0:
        text = text[:i]

    # Now shorten the text
    while len(text) > 70 and p > 60:
        # Cut off 10 chars
        text = "..." + text[10:]
        p = p - 7

    # Now print the string, along with an indicator
    print '> ',text
    print '> ',' '*p + '^'
    print 'List of nearby tokens:', scanner

def wrap_error_reporter(parser, rule):
    try:
        return_value = getattr(parser, rule)()
    except SyntaxError, s:
        input = parser._scanner.input
        try:
            print_error(input, s, parser._scanner)
        except ImportError:
            print 'Syntax Error',s.msg,'on line',1+count(input[:s.pos], '\n')
    except NoMoreTokens:
        print 'Could not complete parsing; stopped around here:'
        print parser._scanner
    return return_value




import sys
from os.path import abspath

error_flag = False
__problems = []

class color:
    '''
        Namespace used for terminal formating.
    '''
    gray = '\033[1;30m'
    red = '\033[1;31m'
    green = '\033[1;32m'
    yellow = '\033[1;33m'
    blue = '\033[1;34m'
    magenta = '\033[1;3m'
    cyan = '\033[1;36m'
    white = '\033[1;37m'
    crimson = '\033[1;38m'

    hred = '\033[1;41m'
    hgreen = '\033[1;42m'
    hbrown = '\033[1;43m'
    hblue = '\033[1;44m'
    hmagenta = '\033[1;45m'
    hcyan = '\033[1;46m'
    hgray = '\033[1;47m'
    hcrimson = '\033[1;48m'

    normal = '\033[1;m'

class Problem (object):
    ''' A problem, with all of its associated information. '''

    def __init__ (self, text, line, error=False):
        global error_flag
        self.text = text
        self.line = line
        if error:
            error_flag = True
        self.error = error

    def display (self, out=sys.stderr):
        if self.line:
            out.write ('%s%s%s:%s%d%s: ' %
                (color.blue, self.line.fname, color.normal,
                 color.green, self.line.linenb, color.normal))
        else:
            out.write ('%s<unk>%s:%s%d%s: ' %
                (color.blue, color.normal, color.green, self.line.linenb, color.normal))
        if self.error:
            out.write ('%serror%s: ' % (color.red, color.normal))
        else:
            out.write ('%swarning%s: ' % (color.yellow, color.normal))
        out.write ('%s\n' % self.text)
        
def warning (text, line=None):
    ''' Adds a warning to the problem list. '''

    __problems.append (Problem (text, line))

def error (text, line=None):
    ''' Adds an error to the problem list. '''

    __problems.append (Problem (text, line, True))

def flush_errors (out=sys.stderr):
    ''' Displays all the errors on the screen and clear the error flag. '''

    for problem in __problems:
        problem.display (out)

    global error_flag
    error_flag = False


__deps = None
def deps_start ():
    ''' Starts the dependancy tracker. '''

    global __deps
    __deps = []

def deps_add (path):
    ''' Add a dependancy to the dependancy list '''

    if __deps is not None:
        __deps.append (abspath (path))

def deps_flush ():
    ''' Empty the list of dependancy and return its contents. '''

    global __deps
    tmp = __deps
    __deps = None
    return tmp




import ply.lex as lex
from common import *

tokens = (
    'ESCAPED',
    'TAG_START',
    # 'PYTHON_SIMPLE',
    # 'PYTHON_COMPLEX',
    'RBRAK',
    'LBRAK',
    'WORD',
    'DQUOTE',
    'SQUOTE',
    'DOT',
    'SHARP',
    'SPACE',
    'DATTR',
    'SATTR',
    'ANYTHING',
)

precedence = (
    ('right', 'TAG_START' ),
    ('right', 'DATTR', 'SATTR'),
    ('left', 'RBRAK', 'LBRAK'),
    ('left', 'DQUOTE', 'SQUOTE'),
    # ('left', 'PYTHON_SIMPLE', 'PYTHON_COMPLEX'),
    ('right', 'ANYTHING', 'WORD', 'SPACE' ), # The lowest
    ('left', 'DOT', 'SHARP'),
    ('left', 'ESCAPED' ),
)

escapes = {
    '<' : '&lt;',
    '>': '&gt;',
}

def t_ESCAPED (t):
    r'\\.'
    if escapes.has_key (t.value[1]):
        t.value = escapes[t.value[1]]
    else:
        t.value = t.value[1:]
    return t

def t_TAG_START (t):
    r'@/?\w[\w0-9_:-]*\s*'
    return t

"""
def t_PYTHON_SIMPLE (t):
    r'\$[\w_][\w0-9_.]*(\((\\.|[^\)])*\))?'
    t.value = t.value[1:].replace ('\\)', ')')
    # FIXME : unescape !
    return t

def t_PYTHON_COMPLEX (t):
    r'\$\s*\{(\\\}|[^\}])*\}'
    start = t.value.find ('{')
    t.value = t.value[start+1:-1].replace ('\\}', '}')
    return t
"""

def t_RBRAK (t):
    r'\]'
    return t
def t_LBRAK (t):
    r'\s*\['
    return t

def t_DQUOTE (t):
    r'"\s*'
    return t

def t_SQUOTE (t):
    r'\'\s*'
    return t

def t_DATTR (t):
    r'\s*\w[\w0-9_\-:]*="'
    return t

def t_SATTR (t):
    r'\s*\w[\w0-9_\-:]*=\''
    return t

def t_WORD (t):
    r'\s*\w[\w0-9_\-:]*\s*'
    return t

def t_DOT (t):
    r'\s*\.\w[\w0-9_:-]*\s*'
    return t

def t_SHARP (t):
    r'\s*\#\w[\w0-9_:-]*\s*'
    return t

def t_ANYTHING (t):
    r'[^\]]'
    return t

def t_error (t):
    print 'ERROR %s' % t.value

lexer = lex.lex ()

##################################### PARSER ####################################


##################################### PWiLang ####################################

from node import *
from os import getcwd, path

current_pwiline = None

def set_current_pwiline (pwiline):
    global current_pwiline
    current_pwiline = pwiline

class Modifiers (object):
    ''' Used to pass parsed modifiers to tags '''
    def __init__ (self):
        self.classes = []
        self.attrs = {}
        self.id = None

    def merge (self, merged):
        if merged.id:
            self.id = merged.id
        self.classes += merged.classes
        self.attrs.update (merged.attrs)
        return self

    def apply (self, tagnode):
        tagnode.classes = self.classes
        tagnode.attrs = self.attrs
        tagnode.id = self.id

def nodify (f, s=None):
    '''
        This function is used to collapse two following ContentNode into
        one (for exemple : 'toto${blublu}tatata' would in the end give three content
        nodes if blublu gives text).
        This is done for cosmetics.
    '''
    res = f

    if isinstance (f, str) or isinstance (f, unicode):
        f = ContentNode (f)
        f.line = current_pwiline
        if not s:
            return f
    else:
        f = f.last # We're only affecting the last node.

    if isinstance (s, str) or isinstance (s, unicode):
        s = ContentNode (s)
        s.line = current_pwiline

    if isinstance (f, ContentNode) and isinstance (s, ContentNode):
        # Merge the two content nodes in one.
        # return ContentNode (f.content + s.content)
        f.content += s.content
        f.last.next = s.next
    else:
        f.last.next = s

    return res

##################################### PLY YACC ####################################
# from pudb import set_trace

import ply.yacc as yacc

def p_top_level (p):
    '''
        top_level : contents
    '''
    p[0] = p[1]

def p_safe_tokens (p):
    '''
        safe_token :  ESCAPED
                    | WORD
                    | SPACE
                    | ANYTHING
    '''
    p[0] = nodify (p[1])

def p_modifiers (p):
    '''
        modifiers : modifiers_rec
                 | modifiers modifiers_rec
    '''
    p[0] = p[1]
    if len (p) == 3:
        p[0].merge (p[2])

def p_modifiers_rec (p):
    '''
        modifiers_rec :  DOT
                       | SHARP
    '''
    p[0] = Modifiers ()
    p[1] = p[1].strip ()
    if p[1].startswith ('.'):
        p[0].classes.append (p[1][1:])
    if p[1].startswith ('#'):
        p[0].id = p[1][1:]

import re
re_escaped = re.compile (r'\\(.)')

def p_tag (p):
    '''
        tag : TAG_START safe_token contents
            | TAG_START tag contents
            | TAG_START tag
            | TAG_START inline_contents
            | TAG_START safe_token
            | TAG_START

        tag_inside : TAG_START safe_token inline_contents_inside
            | TAG_START tag_inside inline_contents_inside
            | TAG_START tag_inside
            | TAG_START inline_contents
            | TAG_START safe_token
            | TAG_START
    '''
    t = TagNode (p[1])
    child = None
    if len (p) == 4:
        child = nodify (p[2], p[3])
    if len (p) == 3:
        if p[2]:
            child = nodify (p[2])
    if len (p[1]) > 1 and p[1][1] == '/':
        t.next = child
    else:
        t.appendChild (child)
    p[0] = t

def p_tag_modif (p):
    '''
        tag : TAG_START modifiers safe_token contents
            | TAG_START modifiers tag contents
            | TAG_START modifiers tag
            | TAG_START modifiers inline_contents
            | TAG_START modifiers safe_token
            | TAG_START modifiers

        tag_inside : TAG_START modifiers safe_token inline_contents_inside
            | TAG_START modifiers tag_inside inline_contents_inside
            | TAG_START modifiers tag_inside
            | TAG_START modifiers inline_contents
            | TAG_START modifiers safe_token
            | TAG_START modifiers
    '''
    t = TagNode (p[1])
    child = None
    mods = p[2]
    if len (p) == 5:
        child = nodify (p[3], p[4])
    if len (p) == 4:
        if p[3]:
            child = nodify (p[3])
    if len (p[1]) > 1 and p[1][1] == '/':
        t.next = child
    else:
        t.appendChild (child)
    p[2].apply (t)
    p[0] = t

def p_inline_contents (p):
    '''
        inline_contents : LBRAK inline_contents_inside RBRAK
                        | LBRAK RBRAK
    '''
    p[0] = None
    if len (p) == 4:
        p[0] = p[2]

def p_modifiers_rec_attrs (p):
    '''
        modifiers_rec : DATTR contents_dattr DQUOTE
                      | SATTR contents_sattr SQUOTE
                      | SATTR SQUOTE
                      | DATTR DQUOTE
    '''
    s = p[1]
    mods = Modifiers ()

    # FIXME : Add the rules
    att = s[:-2]

    if len (p) == 4:
        contents = nodify (p[2])
        contents.prev = nodify (p[3].strip ())
        contents.last.next = nodify (p[3].strip ())
        mods.attrs[att] = contents.prev
    else:
        mods.attrs[att] = nodify ('%s%s' % (p[2], p[2]))
    
    p[0] = mods

def p_contents_recursive (p):
    '''
        contents_rec : safe_token
                      | tag
                      | LBRAK
                      | RBRAK
                      | DOT
                      | SHARP
                      | DATTR
                      | SATTR
                      | DQUOTE
                      | SQUOTE
        inline_contents_inside_rec : safe_token
                                   | tag_inside
                                   | LBRAK
                                   | DOT
                                   | SHARP
                                   | DATTR
                                   | SATTR
                                   | DQUOTE
                                   | SQUOTE
        contents_sattr_rec : safe_token
                          | LBRAK
                          | RBRAK
                          | DOT
                          | SHARP
                          | DATTR
                          | DQUOTE
                          | SATTR
                          | TAG_START
        contents_dattr_rec : safe_token
                          | LBRAK
                          | RBRAK
                          | DOT
                          | SHARP
                          | DATTR
                          | SQUOTE
                          | SATTR
                          | TAG_START
    '''
    p[0] = nodify (p[1])

"""
def p_python_simple (p):
    '''
        safe_token : PYTHON_SIMPLE
                     | PYTHON_COMPLEX
    '''
    p[0] = CommandNode (p[1], current_pwiline)
"""

def p_contents (p):
    '''
        contents : contents_rec
        contents_sattr : contents_sattr_rec
        contents_dattr : contents_dattr_rec
        inline_contents_inside : inline_contents_inside_rec
    '''
    p[0] = nodify (p[1])

def p_contents_2 (p):
    '''
        contents : contents contents_rec
        contents_sattr : contents_sattr contents_sattr_rec
        contents_dattr : contents_dattr contents_dattr_rec
        inline_contents_inside : inline_contents_inside inline_contents_inside_rec
    '''
    p[0] = nodify (p[1], p[2])

def p_error (p):
    import common

    common.error_flag = True
    print current_pwiline
    print 'Error ' + repr (p)

parser = yacc.yacc (debug=None)




#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from BeautifulSoup import BeautifulSoup, NavigableString

import sys

INDENT = "    "

import re
import string
re_not_empty = re.compile ('[^\s]')
re_spaces = re.compile ('\s+', re.MULTILINE)

def print_string_at_indent (s, indent, origindent=""):
    s = s.replace ('[', '\[').replace ('@', '\@').replace (']', '\]')
    lines = s.rstrip (' \n\t').lstrip (' \n\t').split ('\n')
    lines = [lines[0]] + ['\n%s%s' %(indent, l) for l in lines[1:]] + ['\n']
    res = string.join (lines)
    sys.stdout.write (origindent + res)
    
def parse_attrs (node):
    if not node.attrs:
        return ""

    s = ['']
    for key, value in node.attrs:
        if key == "id":
            s.append ('#%s' % value)
        elif key == "class":
            for c in re_spaces.split (value):
                s.append ('.%s' % c)
        else:
            s.append ('%s="%s"' % (key, value))
    return string.join (s, ' ')

def treat_node (node, indent=""):
    # if node.string -- single content node.

    if not isinstance (node, NavigableString):
        sys.stdout.write ('%s@%s' % (indent, node.name))
        sys.stdout.write (parse_attrs (node))
        if node.string:
            sys.stdout.write (' ')
            print_string_at_indent (' %s' % (str (node.string)), indent + INDENT)
        else:
            sys.stdout.write ('\n')
            for n in node.contents:
                treat_node (n, indent + INDENT)
    else:
        if re_not_empty.search (str (node)):
            print_string_at_indent (str (node), indent, indent)

def treat_file (fname):
    f = open (fname, 'r')
    contents = f.read ()
    contents = contents.decode ('utf-8')
    f.close ()

    soup = BeautifulSoup (contents)
    for n in soup.contents:
        treat_node (n, "")

for fname in sys.argv[1:]:
    treat_file (fname)



#!/usr/bin/env python

import re
import sys

from os import path

default_dirs = [path.expanduser("~/.pwilang/templates")]

from common import error, warning, deps_add

re_comment = re.compile ("^\s*#.*$")

############### PwiLines #########

re_indent = re.compile (r'(\s*)(.*)', re.DOTALL)
class PwiLine (object):
    '''
        This class represents a single line in the pwilang world.
        It possesses the following information :
        The line number in the original file, the original file, the indentation of the line, and of course its contents.
    '''

    def __init__ (self, content, linenb, fname):
        self.linenb = linenb
        self.fname = fname

        m = re_indent.match (content)
        self.indent = len (m.group(1))
        self.content = m.group(2)

    def __repr__ (self):
        return '%d:%s' % (self.linenb, self.content)

############# MACROS ####################

def warning (text):
    sys.stderr.write ("%s\n" % text)

def real_lines (lines_arr):
    '''
        Sorts out comments, empty lines and multi lines. 
        Acts as a generator.
    '''
    raw_block = False
    nb = 0

    in_multiline = False
    acc = ""

    for line in lines_arr:
        nb += 1
        line = line.rstrip ()

        if line.endswith ('\\'):
            acc += '%s\n' % (line[:-1]) # Removing the \
            continue
        if line == '': # discarding empty lines
            continue
        if line.startswith ('#'): # comment
            continue
        if acc:
            acc += line
            yield acc, nb
            acc = ""
            continue
        yield line, nb
    if acc:
        yield acc, nb

def real_lines_getter (iterative, fname="<str>"):
    lines = []
    for l, nb in real_lines (iterative):
        lines.append (PwiLine (l, nb, fname))
    return lines

def get_lines_from_iterable (iterable, file_like=True):
    return real_lines_getter (iterable)

def get_lines_from_file (fname):
    ''' Get the complete lines from a file (with blocks resolved and includes executed) '''
    f = open (fname, 'r')
    lines = get_lines_from_iterable (f, fname)
    f.close ()
    return lines

def get_lines_from_string (str):
    lines = real_lines_getter (str.split ('\n'))
    return lines






class NodeIterator (object):
    ''' An iterator that iterates over a node's level. '''

    def __init__ (self, node, forward=True, real=False):
        self.__node = node
        self.__forward = forward
        self.__real = real

    def next (self):
        res = self.__node
        if not res:
            raise StopIteration
        if self.__forward:
            if self.__real:
                self.__node = self.__node.realnext
            else:
                self.__node = self.__node.next
        else:
            if self.__real:
                self.__node = self.__node.realprev
            else:
                self.__node = self.__node.prev
        return res

    def __iter__ (self):
        return self

class NodeLineIterator (object):
    ''' An iterator that iterates over a 'line' of nodes (see Node#lastInLine documentation) '''

    def __init__ (self, node, forward=True, real=False):
        self.__first = node
        self.__node = node
        self.__forward = forward
        self.__real = real

    def next (self):
        res = self.__node
        if not res or self.__node.indent != -1 and self.__node != self.__first:
            raise StopIteration
        if self.__forward:
            if self.__real:
                self.__node = self.__node.realnext
            else:
                self.__node = self.__node.next
        else:
            if self.__real:
                self.__node = self.__node.realprev
            else:
                self.__node = self.__node.prev
        return res

    def __iter__ (self):
        return self

class NodeDepthIterator (object):
    ''' 
        An iterator that recurses over the whole Node tree in depth-first order.

        It provides two useful methods : remove () and replace () that are to be used
        instead of Node#pop () and Node#replace () since these would most likely disrupt
        the iterator's process.
    '''

    def __init__ (self, node):
        self.__current_node = None
        self.__next_node = node

    def __compute_next (self):
        ''' Compute the next node in the iteration '''
        if not self.__current_node:
            return
        if self.__current_node.child:
            self.__next_node = self.__current_node.child
        elif self.__current_node.next:
            self.__next_node = self.__current_node.next
        else: # We have to backtrack
            tmp = self.__current_node.parent
            while tmp and not tmp.next:
                tmp = tmp.parent
            if tmp:
                self.__next_node = tmp.next
            else:
                self.__next_node = None

    def next (self):
        self.__compute_next ()
        if not self.__next_node:
            raise StopIteration ()
        self.__current_node = self.__next_node
        return self.__current_node

    def remove (self):
        ''' Removes the current node from the tree without affecting the iterator. '''
        if not self.__current_node:
            raise ValueError ('There is no current node')
        prev = self.__current_node.prev
        parent = self.__current_node.parent
        next = self.__current_node.pop ()

        # We first try to come back to previous nodes so that __compute_next ()
        # will go to the next node by itself.
        if prev:
            self.__current_node = prev
        elif parent: # The removed node was the first child on its level.
            self.__current_node = parent
        elif next: # The node was the first one at root level.
            self.__current_node = None
            self.__next_node = next
        else: # The tree is now empty.
            self.__current_node = None
            self.__next_node = None

    def replace (self, new):
        ''' Replace the current node (the last returned by next()) by a new one in the tree.
            The next node in the iteration will be the child (or the next) of the new one. '''
        if not self.__current_node:
            raise ValueError ('There is no current node')
        new.replace (self.__current_node)
        self.__current_node = new

    def __iter__ (self):
        return self

class Node (object):
    '''
        A Node tree is the intermediary data structure resulting from the parsing of a .pwi file.
        It consists of nodes that hold data (such as a html tag, content, or python code) that point to
        either, both or no child and next node.
    '''

    def __init__ (self):
        self.parent = None
        self.__next = None
        self.__prev = None
        self.__child = None

        # Here we cache the last brother to avoid looking it up everytime.
        self.__last = None 
        # Same for first.
        self.__first = None

        self.indent = -1
        self.line = None

    def __invalidateCache (self):
        '''
            When the tree structure changes in a significant way, we have to cancel the cache
            to avoid spurious .last or .first calls.
        '''
        self.__last = None
        self.__first = None

    # Pops an element from the siblings.
    def pop (self):
        '''
            Pops a node and its children out of its tree. This method sets the node's next 
            and prev to None in addition to removing it from the tree (updating its previous, 
            parent and next node if need be).
            Returns the next element (or None if there was no next)
        '''
        _next = self.__next
        _prev = self.__prev

        if _next:
            _next.__prev = _prev
        if _prev:
            _prev.__next = _next

        if self.parent and self.parent.child == self:
            self.parent.child = _next

        self.__next = None
        self.__prev = None
        self.parent = None
        
        self.__invalidateCache ()

        return _next
        
    def fixindent (self, indent, modifier=4):
        '''
            Fixes the indentation of a node to the 'indent' parameter.
            The modifier variable is for indentation incrementation for the children nodes.
            This function should be used whenever the tree structure changes in a significant
            way to ensure that the nodes stay correctly indented.
            Attention : The indentation doesn't affect how the tree is parsed. This is just
            for cosmetics and the pretty (indented) output.
        '''
        for sibling in self:
            if sibling.child:
                if indent == -1:
                    sibling.child.fixindent (indent, modifier)
                else:
                    sibling.child.fixindent (indent + modifier, modifier)
            if sibling.indent != -1: # inline nodes should not be affected.
                sibling.indent = indent

    def replace (self, node):
        '''
            Replaces the 'node' in its tree. The correct syntax is when one wants to replace 
            and 'old' node by a 'new' one : new.replace (old).
            The new node inserts its whole tree structure (children and following brothers 
            -- not previous ones).
        '''
        # prev, next, parent and parent.child are affected.
        if not node:
            return # We can't replace something that doesn't exist

        l = self.last
        l.__next = node.__next
        if node.__next:
            node.__next.__prev = l
        self.__prev = node.__prev
        if node.__prev:
            node.__prev.__next = self

        self.__invalidateCache ()

        self.parent = node.parent
        if node.parent and node.parent.child == node:
            node.parent.child = self

        for n in self:
            n.parent = node.parent

        # Some cleanup. Since the node is not part of the tree, it shouldn't have
        # dangling pointers.
        node.__next = None
        node.__prev = None
        node.__invalidateCache ()

    def appendChild (self, node):
        '''
            Appends a node to the children of this node.
            If there weren't any then the new node becomes the only child.
        '''
        if self.child:
            node.parent = self
            self.child.last.next = node
        else:
            self.child = node

    ###################################### PROPERTIES #######################################

    def _next_get (self):
        ''' The next node on the same level. '''
        return self.__next

    def _next_set (self, val):
        self.__invalidateCache ()

        if self.__next:
            self.__next.__prev = None

        if val:
            _val_last = val.last

            for n in val:
                n.parent = self.parent

            val.__prev = self
            _val_last.__next = self.__next

            if self.__next:
                self.__next.__prev = _val_last
        self.__next = val
    next = property (fget=_next_get, fset=_next_set)

    def _prev_get (self):
        ''' The previous node on the same level '''
        return self.__prev

    def _prev_set (self, val):
        self.__invalidateCache ()
        if self.__prev:
            self.__prev.next = None
        if val:
            for n in val:
                n.parent = self.parent
            val.__next = self
            val.__prev = self.__prev
            if self.__prev:
                self.__prev.__next = val
            if self.parent and self.parent.child == self:
                self.parent.child = val
        else:
            if self.parent:
                self.parent.child = self
        self.__prev = val
    prev = property (fget=_prev_get, fset=_prev_set)

    def _child_get (self):
        ''' The first child of the node '''
        return self.__child

    def _child_set (self, val):
        if val:
            for n in val:
                n.parent = self
        self.__child = val
    child = property (fget=_child_get, fset=_child_set)

    def _last_get (self):
        ''' The last brother of the node (its next of next of next...) '''
        if self.__last and not self.__last.next:
            # The last is already cached
            return self.__last
        cur = self.__last
        if not cur: # There was no cache
            cur = self
        while cur.next:
            cur = cur.next
        self.__last = cur
        return cur
    last = property (fget=_last_get)

    def _lastInLine_get (self):
        ''' 
            The last brother of the node in the same line.
            A node is on the same line of another if in the source .pwi file
            the source code that led to their creation was on the same line.
            Ex: @p {toto} @p {titi}
            creates two 'p' nodes that are on the same line.
            In practice, a line always starts with a node with its .indent attribute
            >= 0, and subsequent nodes with a -1 indentation.
        '''
        tmp = self
        for n in self.lineIter ():
            tmp = n
        return tmp
    lastInLine = property (fset=_lastInLine_get)

    def _first_get (self):
        '''
            The first node on the same level.
        '''
        if self.__first and not self.__first.prev:
            return self.__first # Cached
        cur = self.__first
        if not cur:
            cur = self
        while cur.prev:
            cur = cur.prev
        self.__first = cur
        return cur
    first = property (fset=_first_get)

    ############################### ITERATORS #############################

    def __iter__ (self):
        ''' An iterator on the node's sibling (including itself) '''
        return NodeIterator (self)

    def lineIter (self):
        ''' An iterator on the node's sibling that are on the same line (including itself) '''
        return NodeLineIterator (self)

    def forward (self):
        ''' An iterator on the node's sibling (including itself) '''
        return NodeIterator (self)

    def __reversed__ (self):
        ''' An iterator on the node's sibling (including itself) in reverse order. '''
        return NodeIterator (self, False)
    def reverse (self):
        ''' An iterator on the node's sibling (including itself) in reverse order. '''
        return NodeIterator (self, False)

    def treeIter (self):
        ''' An iterator on the node's whole tree in depth-first order (including itself) '''
        return NodeDepthIterator (self)


    ############################### CLONE #################################

    def clone (self):
        ''' Clones the current node and its children (deep copy) '''
        raise NotImplemented () # Has to be rederivated.

    def cloneChildren (self):
        ''' Clone the children of the current node (deep copy) '''
        children = EmptyNode ()
        if self.child:
            for c in self.child:
                children.last.next = c.clone ()
        return children.pop ()


class TagNode (Node):
    ''' A node holding an XML tag '''

    def __init__ (self, tag):
        Node.__init__ (self)
        if tag.startswith ('@'):# Removing the @
            tag = tag[1:]
        tag = tag.strip () 
        self.tag = tag

        self.self_closing = False
        if tag.startswith ('/'):
            self.tag = tag[1:]
            self.self_closing = True

        self.classes = []
        self.id = None
        self.attrs = {}
        self.inline = False

    def __repr__ (self):
        return "%s@%s" % (self.tag, id(self))

    def clone (self):
        ''' Clones the current node '''
        n = TagNode (self.tag)
        n.classes = self.classes
        n.id = self.id
        n.attrs = self.attrs
        n.child = self.cloneChildren ()
        n.indent = self.indent
        return n

class ContentNode (Node):
    ''' A node that holds text '''

    def __init__ (self, content):
        Node.__init__ (self)
        self.content = content
    def __repr__ (self):
        return "'%s'@%s" % (self.content, id(self))

    def clone (self):
        ''' Clones the current node '''
        n = ContentNode (self.content)
        n.child = self.cloneChildren ()
        n.indent = self.indent
        return n

class EmptyNode (Node):
    """ A Node that is empty. """
    def __init__ (self):
        Node.__init__ (self)
    def __repr__ (self):
        return "  --@%s" % (id(self))

    def clone (self):
        ''' Clones the current node '''
        n = EmptyNode ()
        n.child = self.cloneChildren ()
        n.indent = self.indent
        return n




from jinja2 import nodes, compiler, parser
from jinja2.ext import Extension
from jinja2 import environmentfilter, Markup, contextfilter, environmentfunction

from os.path import abspath, dirname, join
from os import getcwd

#@environmentfilter
#def abspath (value, env):
#    pass

filters = {
}

import types

def can_assign (self):
    return True
setattr (nodes.Getitem, 'can_assign', can_assign)
setattr (nodes.Getattr, 'can_assign', can_assign)

def parse_assign_target(self, with_tuple=True, name_only=False,
                        extra_end_rules=None):
    """Parse an assignment target.  As Jinja2 allows assignments to
    tuples, this function can parse all allowed assignment targets.  Per
    default assignments to tuples are parsed, that can be disable however
    by setting `with_tuple` to `False`.  If only assignments to names are
    wanted `name_only` can be set to `True`.  The `extra_end_rules`
    parameter is forwarded to the tuple parsing function.
    """
    if self.stream.look ().value == 'in':
        name_only = True

    if name_only:
        token = self.stream.expect('name')
        target = nodes.Name(token.value, 'store', lineno=token.lineno)
    else:
        if with_tuple:
            target = self.parse_tuple(simplified=False,
                                      extra_end_rules=extra_end_rules)
        else:
            target = self.parse_primary()
        target.set_ctx('store')
    if not target.can_assign():
        print target
        self.fail('can\'t assign to %r' % target.__class__.
                  __name__.lower(), target.lineno)
    return target

setattr (parser.Parser, 'parse_assign_target', parse_assign_target)

def visit_Getattr(self, node, frame):
    if node.ctx == "load":
        self.write('environment.getattr(')
        self.visit(node.node, frame)
        self.write(', %r)' % node.attr)
    else:
        self.visit(node.node, frame)
        self.write ('.')
        self.write(node.attr)

setattr (compiler.CodeGenerator, 'visit_Getattr', visit_Getattr)

def visit_Getitem(self, node, frame):
    # slices bypass the environment getitem method.
    # We bybasse the environment getitem method since it's pretty useless anyway !
    # if isinstance(node.arg, nodes.Slice):
    self.visit(node.node, frame)
    self.write('[')
    self.visit(node.arg, frame)
    self.write(']')
   # else:
   #     self.write('environment.getitem(')
   #     self.visit(node.node, frame)
   #     self.write(', ')
   #     self.visit(node.arg, frame)
   #     self.write(')')
setattr (compiler.CodeGenerator, 'visit_Getitem', visit_Getitem)


class AbspathExtension (Extension):
    tags = set (['abspath'])

    def __init__ (self, env):
        self.__env = env
        super (AbspathExtension, self).__init__ (env)

    def parse (self, parser):
        filename = parser.filename

        if not filename:
            dir = getcwd ()
        else:
            dir = dirname (parser.filename)

        tag = next (parser.stream) # Removing from the stream
            # the "abspath" tag

        filename = parser.stream.expect ("string")

        while parser.stream.current.type != "block_end":
            pass # We iterate until the end.

        path = abspath (join (dir, filename.value))

        return nodes.Output ([nodes.Const (unicode (path))])

@environmentfunction
def set_global (env, name, value):
    env.globals[name] = value
    return ""

@environmentfunction
def get_global (env, name):
    return env.globals.get (name, None)




import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "pwilang",
    version = "0.10.6",
    author = "Christophe Eymard",
    author_email = "christophe@ravelsoft.com",
    description = ("A text processor with its own syntax a little nicer than html that uses jinja for complex text rendering."),
    license = "GPLv3",
    keywords = "text processor html jinja",
    url = "http://code.google.com/p/pwilang/",
    packages=['pwilang'],
    scripts=["pwilang/pwilang"],
    long_description=read('README'),
    install_requires=['jinja2>=2.4', 'ply>=3.1', 'beautifulsoup'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
    ],
)

